from setuptools import setup

setup(
    name='carnotpy',
    version='0.10.4',
    author='Carnot Innovations',
    description='Carnot Connection as a Python Package',
    long_description='Carnot Connection as a Python Package',
    py_modules=['carnotpy'],
    install_requires=[
        'pandas',
        'requests',
        'python-dateutil',
    ],
)



import pandas as pd
import datetime
import requests
from dateutil.relativedelta import relativedelta
import json


def process_equipment(payload, equipment_name):
    if payload['filters'].get('equipment_name'):
        if payload.get('pipeline') and len(payload['pipeline']) > 0:
            for stage in payload['pipeline']:
                if stage['type'] == 'filter':
                    stage['filter_by'] = stage['filter_by'].replace(payload['filters']['equipment_name'] + ' - ',
                                                                    equipment_name + ' - ')
                elif stage.get('formula') and stage['formula'] != '':
                    stage['formula'] = stage['formula'].replace(payload['filters']['equipment_name'] + ' - ',
                                                                equipment_name + ' - ')
                elif stage['type'] == 'insert_series':
                    stage['name'] = stage['name'].replace(payload['filters']['equipment_name'] + ' - ',
                                                          equipment_name + ' - ')
                    if stage.get('filters') and stage['filters']['equipment_name'] == payload['filters'][
                        'equipment_name']:
                        stage['filters']['equipment_name'] = equipment_name
        payload['filters']['equipment_name'] = equipment_name


def process_date_range(payload, date_range):
    if payload.get('useCustomRange'):
        if payload.get('useCustomRelativeRange'):
            rel_range = payload.get('customRelativeRange', '30 days')
            new_range = None
            base_date = datetime.datetime.now()
            start_date = base_date - datetime.timedelta(days=30)
            if rel_range == 'current week':
                start_date = base_date - datetime.timedelta(days=base_date.weekday())
            elif rel_range == 'current month':
                start_date = base_date.replace(day=1)
            elif rel_range == 'today' or rel_range == "today full":
                start_date = base_date.replace(hour=0, minute=0, second=0, microsecond=0)
            elif rel_range == 'current year':
                start_date = base_date.replace(month=1, day=1)
            elif rel_range == 'last week':
                base_date = base_date - datetime.timedelta(weeks=1)
                end_date = base_date.replace(hour=23, minute=59, second=59, microsecond=999999)
                start_date = end_date - datetime.timedelta(days=end_date.weekday())
            elif rel_range == 'last month':
                base_date = base_date.replace(day=1) - datetime.timedelta(days=1)
                end_date = base_date.replace(hour=23, minute=59, second=59, microsecond=999999)
                start_date = end_date.replace(day=1)
            elif rel_range == 'yesterday':
                base_date = base_date - datetime.timedelta(days=1)
                end_date = base_date.replace(hour=23, minute=59, second=59, microsecond=999999)
                start_date = end_date.replace(hour=0, minute=0, second=0, microsecond=0)
            elif rel_range == 'last year':
                base_date = base_date.replace(year=base_date.year - 1, month=12, day=31)
                end_date = base_date.replace(hour=23, minute=59, second=59, microsecond=999999)
                start_date = end_date.replace(year=base_date.year - 1, month=1, day=1)
            elif rel_range == 'today and tomorrow':
                start_date = base_date.replace(hour=0, minute=0, second=0, microsecond=0)
                end_date = base_date + datetime.timedelta(days=1)
                base_date = end_date.replace(hour=23, minute=59, second=59, microsecond=999999)

            if rel_range in ['yesterday', 'today', 'today and tomorrow']:
                new_range = {
                    'start': start_date.strftime('%Y-%m-%d %H:%M:%S'),
                    'end': base_date.strftime('%Y-%m-%d %H:%M:%S')
                }
            else:
                new_range = {
                    'start': start_date.strftime('%Y-%m-%d'),
                    'end': (base_date.strftime('%Y-%m-%d %H:%M:%S') if len(
                        base_date.strftime('%Y-%m-%d')) < 12 else base_date.strftime('%Y-%m-%d'))
                }
            if payload.get('customRelativeRangeStartOnly') and date_range and date_range['end']:
                payload['date_range'] = [new_range['start'], (
                    date_range['end'] + ' 23:59:59' if len(date_range['end']) < 12 else date_range['end'])]
            else:
                payload['date_range'] = [new_range['start'], (
                    new_range['end'] + ' 23:59:59' if len(new_range['end']) < 12 else new_range['end'])]

        elif payload.get('customRange'):
            if payload.get('customRangeStartOnly') and date_range and date_range['end']:
                payload['date_range'] = [payload['customRange']['start'], (
                    date_range['end'] + ' 23:59:59' if len(date_range['end']) < 12 else date_range['end'])]
            else:
                payload['date_range'] = [payload['customRange']['start'], (
                    payload['customRange']['end'] + ' 23:59:59' if len(payload['customRange']['end']) < 12 else
                    payload['customRange']['end'])]

    elif date_range:
        if payload.get('offsetRange') and payload.get('offsetType'):
            if payload['offsetRangeStartOnly']:
                if payload['offsetType'] in ['year', 'years'] and payload.get('offsetPreciseYear'):
                    payload['date_range'] = [
                        datetime.datetime.strptime(date_range['start'], '%Y-%m-%d').replace(
                            year=payload['offsetPreciseYear']).strftime('%Y-%m-%d'),
                        (date_range['end'] + ' 23:59:59' if len(date_range['end']) < 12 else date_range['end'])
                    ]
                elif payload['offsetType'] in ['month', 'months']:
                    offset = relativedelta(months=-payload['offsetAmount'])
                    payload['date_range'] = [
                        (datetime.datetime.strptime(date_range['start'], '%Y-%m-%d') + offset).strftime('%Y-%m-%d'),
                        (date_range['end'] + ' 23:59:59' if len(date_range['end']) < 12 else date_range['end'])
                    ]
                elif payload['offsetType'] in ['year', 'years']:
                    offset = relativedelta(years=-payload['offsetAmount'])
                    payload['date_range'] = [
                        (datetime.datetime.strptime(date_range['start'], '%Y-%m-%d') + offset).strftime('%Y-%m-%d'),
                        (date_range['end'] + ' 23:59:59' if len(date_range['end']) < 12 else date_range['end'])
                    ]
                else:
                    payload['date_range'] = [
                        (datetime.datetime.strptime(date_range['start'], '%Y-%m-%d') - datetime.timedelta(
                            **{payload['offsetType']: payload['offsetAmount']})).strftime('%Y-%m-%d'),
                        (date_range['end'] + ' 23:59:59' if len(date_range['end']) < 12 else date_range['end'])
                    ]
            else:
                if payload['offsetType'] in ['year', 'years'] and payload.get('offsetPreciseYear'):
                    payload['date_range'] = [
                        datetime.datetime.strptime(date_range['start'], '%Y-%m-%d').replace(
                            year=payload['offsetPreciseYear']).strftime('%Y-%m-%d'),
                        datetime.datetime.strptime(date_range['end'], '%Y-%m-%d').replace(
                            year=payload['offsetPreciseYear']).strftime('%Y-%m-%d') + ' 23:59:59'
                    ]
                elif payload['offsetType'] in ['month', 'months']:
                    offset = relativedelta(months=-payload['offsetAmount'])
                    payload['date_range'] = [
                        (datetime.datetime.strptime(date_range['start'], '%Y-%m-%d') + offset).strftime('%Y-%m-%d'),
                        (datetime.datetime.strptime(date_range['end'], '%Y-%m-%d') + offset).strftime(
                            '%Y-%m-%d') + ' 23:59:59'
                    ]
                elif payload['offsetType'] in ['year', 'years']:
                    offset = relativedelta(years=-payload['offsetAmount'])
                    payload['date_range'] = [
                        (datetime.datetime.strptime(date_range['start'], '%Y-%m-%d') + offset).strftime('%Y-%m-%d'),
                        (datetime.datetime.strptime(date_range['end'], '%Y-%m-%d') + offset).strftime(
                            '%Y-%m-%d') + ' 23:59:59'
                    ]
                else:
                    payload['date_range'] = [
                        (datetime.datetime.strptime(date_range['start'], '%Y-%m-%d') - datetime.timedelta(
                            **{payload['offsetType']: payload['offsetAmount']})).strftime('%Y-%m-%d'),
                        (datetime.datetime.strptime(date_range['end'], '%Y-%m-%d') - datetime.timedelta(
                            **{payload['offsetType']: payload['offsetAmount']})).strftime('%Y-%m-%d') + ' 23:59:59'
                    ]
        else:
            payload['date_range'] = [date_range['start'], (
                date_range['end'] + ' 23:59:59' if len(date_range['end']) < 12 else date_range['end'])]


class CarnotPy:

    def __init__(self, url, username, password, preset_token=None):
        self.url = url
        self.username = username
        self.password = password
        self.preset_token = preset_token
        if preset_token:
            self.refresh_token(long_duration=True)

    def get_token(self):
        if self.preset_token is not None:
            return self.preset_token
        result = requests.post(self.url + "/login/", json={"user_name": self.username, "password": self.password})
        assert result.status_code == 200
        try:
            token = result.json()["token"]
            return token
        except KeyError:
            raise ValueError("Invalid username or password")

    def refresh_token(self, long_duration=False):
        """
        Refresh the current JWT token via the /refresh-token/ endpoint.
        Optionally request a long-duration token.

        :param long_duration: If True, request a token with extended expiration
        :return: The new token string
        """
        result = requests.post(
            self.url + "/refresh-token/",
            json={"long_expiration": long_duration},
            headers={"Authorization": "Bearer " + self.preset_token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        data = result.json()
        if data.get('result') != 'OK':
            raise ValueError(data.get('desc', 'Token refresh failed'))
        new_token = data['token']
        self.preset_token = new_token
        return new_token

    def discover(self):
        """
        Get the point list of the facility
        :return: return a point list
        """
        # return a point list
        token = self.get_token()
        result = requests.post(self.url + "/point-discovery/",
                               headers={"Authorization": "Bearer " + token}
                               )
        assert result.status_code == 200
        return result.json()["result"]

    def get_notifications(self, facility_name, date_range=None, ignore_acknowledged=True):
        token = self.get_token()
        payload={
            "facility_name": facility_name,
            "date_range": date_range,
            "ignore_acknowledged": ignore_acknowledged
        }
        result = requests.post(
            self.url + "/notification/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    def get_config(self, facility_name):
        token = self.get_token()
        payload={
            "facility_name": facility_name,
        }
        result = requests.post(
            self.url + "/config/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    def set_config(self, facility_name, config_name, config, config_id=None):
        token = self.get_token()
        result = requests.put(
            self.url + "/config/",
            json={
                "facility_name": facility_name,
                "config_name": config_name,
                "config": config,
                "config_id": config_id
            },
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def get_maintenance_records(self, facility_name, date_range=None, equipment_name=None):
        token = self.get_token()
        payload={
            "facility_name": facility_name,
            "date_range": date_range,
            "equipment_name": equipment_name
        }
        result = requests.post(
            self.url + "/maintenancerecord/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    def read_history(self, sensor_ids, start_time, end_time):
        token = self.get_token()
        result = requests.post(
            self.url + "/historical-data/",
            json={
                "date_range": [start_time, end_time],
                "sensor_ids": sensor_ids
            },
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        data = result.json()
        data['timestamp'] = pd.to_datetime(data['timestamp'])
        df = pd.DataFrame(data)
        df = df.set_index('timestamp')
        return df

    def read_data3(self, payload, token=None):
        if token is None:
            token = self.get_token()
        result = requests.post(
            self.url + "/data3/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        data = json.loads(result.text)  # check if it's valid json, will raise an error if not
        # data = result.json()
        data['timestamp'] = pd.to_datetime(data['timestamp'])
        df = pd.DataFrame(data)
        df = df.set_index('timestamp')
        return df

    def read_data3_query(self, query, facility_name, date_range, equipment=None, on_only=False, parameters=None, data3_implementation=None, independent_x_axis=False):
        if data3_implementation is None:
            def get_read_data3_func():
                # return a function that reads data3 with the token already
                token = self.get_token()
                def my_read_data3(payload):
                    return self.read_data3(payload, token)
                return my_read_data3
            data3_implementation = get_read_data3_func()
        payload_array = query['data-payload']
        df = None
        index = None
        for payload in payload_array:
            payload['parameters'] = parameters
            process_date_range(payload, date_range)
            if 'facility_name' not in payload["filters"]:
                payload["filters"]["facility_name"] = facility_name
            if equipment is not None:
                process_equipment(payload, equipment)
            if on_only or payload.get('onOnly', False):
                if 'pipeline' not in payload:
                    payload['pipeline'] = []
                payload['pipeline'].append({
                                  "type": "filter",
                                  "filter_op": "eq",
                                  "filter_by": "* - on",
                                  "filter_value": 1,
                              })
            del(payload["seriesOverrides"])
            new_df = data3_implementation(payload)
            if index is None:
                index = new_df.index
            if independent_x_axis:
                if df is None:
                    df = new_df
                else:
                    df = pd.concat([df, new_df], axis=1)
            else:
                new_df.index = range(len(new_df))
                if df is None:
                    df = new_df
                else:
                    df = df.join(new_df)
        if not independent_x_axis and df is not None:
            df.index = index
        return df

    def get_sensor_id(self, facility_name, equipment_name, sensor_name):
        token = self.get_token()
        result = requests.post(self.url + "/point/",
                               json={"facility_name": facility_name, "equipment_name": equipment_name},
                               headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
                               )
        assert result.status_code == 200
        points = result.json()['result']
        for point in points:
            if point['sensor_name'] == sensor_name:
                return point['sensor_id']
        raise KeyError("Sensor not found")

    def write(self, sensor_id, value, priority, duration=None):
        token = self.get_token()
        result = requests.put(
            self.url + "/optimizationcontrol/",
            json={
                "sensor_id": sensor_id,
                "value": value,
                "priority": priority,
                "duration": duration
            },
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def add_job(self, facility_name, job_type, meta):
        token = self.get_token()
        result = requests.put(
            self.url + "/job/",
            json={
                "facility_name": facility_name,
                "job_type": job_type,
                "job_status": "pending",
                "meta": meta,
            },
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    # ------------------------------------------------------------------ #
    #  Point Management (CRUD)                                            #
    # ------------------------------------------------------------------ #

    def get_points(self, facility_name, equipment_name=None):
        """
        Get the full list of points (sensors) for a facility, optionally
        filtered by equipment.  Returns complete records including
        calculations_statement, unit, meta, connector_id, etc.

        :param facility_name: Name of the facility
        :param equipment_name: Optional equipment name filter
        :return: list of point dicts
        """
        token = self.get_token()
        payload = {"facility_name": facility_name}
        if equipment_name is not None:
            payload["equipment_name"] = equipment_name
        result = requests.post(
            self.url + "/point/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    def create_point(self, facility_name, equipment_name, sensor_name,
                     sensor_name_abbr=None, sensor_desc=None, unit=None,
                     haystack_id=None, calculations_statement=None,
                     weight=None, connector_id=None, faulty=False, meta=None):
        """
        Create a new point (sensor) under the given facility / equipment.

        :param facility_name: Facility the point belongs to
        :param equipment_name: Equipment the point belongs to
        :param sensor_name: Display name of the sensor
        :param sensor_name_abbr: Abbreviated sensor name
        :param sensor_desc: Description
        :param unit: Measurement unit (e.g. "degC", "kW")
        :param haystack_id: External system haystack ID
        :param calculations_statement: Calculation formula, e.g. "df[2]+df[3]"
        :param weight: Sensor weight (default 1)
        :param connector_id: ID of the connector providing data
        :param faulty: Mark sensor as faulty
        :param meta: JSON metadata dict (tags, writable_priority_array, etc.)
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "equipment_name": equipment_name,
            "sensor_name": sensor_name,
            "faulty": faulty,
        }
        if sensor_name_abbr is not None:
            payload["sensor_name_abbr"] = sensor_name_abbr
        if sensor_desc is not None:
            payload["sensor_desc"] = sensor_desc
        if unit is not None:
            payload["unit"] = unit
        if haystack_id is not None:
            payload["haystack_id"] = haystack_id
        if calculations_statement is not None:
            payload["calculations_statement"] = calculations_statement
        if weight is not None:
            payload["weight"] = weight
        if connector_id is not None:
            payload["connector_id"] = connector_id
        if meta is not None:
            payload["meta"] = meta
        result = requests.put(
            self.url + "/point/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def update_point(self, sensor_id, facility_name, equipment_name, **kwargs):
        """
        Update an existing point (sensor).

        Supported kwargs: sensor_name, sensor_name_abbr, sensor_desc, unit,
        haystack_id, calculations_statement, weight, connector_id, faulty, meta.

        :param sensor_id: The sensor_id to update
        :param facility_name: Facility the point belongs to
        :param equipment_name: Equipment the point belongs to
        :param kwargs: Fields to update
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "equipment_name": equipment_name,
            "sensor_id": sensor_id,
        }
        allowed_fields = [
            'sensor_name', 'sensor_name_abbr', 'sensor_desc', 'unit',
            'haystack_id', 'calculations_statement', 'weight',
            'connector_id', 'faulty', 'meta'
        ]
        for key, value in kwargs.items():
            if key in allowed_fields:
                payload[key] = value
        result = requests.put(
            self.url + "/point/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def delete_point(self, sensor_id):
        """
        Delete a point (sensor) and all its readings.

        :param sensor_id: The sensor_id to delete
        :return: dict with 'result'
        """
        token = self.get_token()
        result = requests.delete(
            self.url + "/point/",
            json={"sensor_id": sensor_id},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    # ------------------------------------------------------------------ #
    #  Equipment Management (CRUD)                                        #
    # ------------------------------------------------------------------ #

    def get_equipment(self, facility_name):
        """
        List all equipment for a facility.

        :param facility_name: Name of the facility
        :return: list of equipment dicts
        """
        token = self.get_token()
        result = requests.post(
            self.url + "/equipment/",
            json={"facility_name": facility_name},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    def create_equipment(self, facility_name, equipment_name,
                         equipment_type=None, equipment_tags=None,
                         floor=None, notes=None, meta=None, graphic_id=None):
        """
        Create a new equipment under the given facility.

        :param facility_name: Facility name
        :param equipment_name: Equipment name
        :param equipment_type: Type of equipment (e.g. "chiller", "ahu")
        :param equipment_tags: Tags list
        :param floor: Floor identifier
        :param notes: Free-text notes
        :param meta: JSON metadata dict
        :param graphic_id: Associated graphic ID
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "equipment_name": equipment_name,
        }
        if equipment_type is not None:
            payload["equipment_type"] = equipment_type
        if equipment_tags is not None:
            payload["equipment_tags"] = equipment_tags
        if floor is not None:
            payload["floor"] = floor
        if notes is not None:
            payload["notes"] = notes
        if meta is not None:
            payload["meta"] = meta
        if graphic_id is not None:
            payload["graphic_id"] = graphic_id
        result = requests.put(
            self.url + "/equipment/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def update_equipment(self, equipment_id, facility_name, equipment_name, **kwargs):
        """
        Update an existing equipment.

        Supported kwargs: equipment_type, equipment_tags, floor, notes, meta,
        graphic_id.

        :param equipment_id: The equipment_id to update
        :param facility_name: Facility the equipment belongs to
        :param equipment_name: New or existing equipment name
        :param kwargs: Fields to update
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "equipment_name": equipment_name,
            "equipment_id": equipment_id,
        }
        allowed_fields = [
            'equipment_type', 'equipment_tags', 'floor', 'notes', 'meta',
            'graphic_id'
        ]
        for key, value in kwargs.items():
            if key in allowed_fields:
                payload[key] = value
        result = requests.put(
            self.url + "/equipment/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def delete_equipment(self, equipment_id):
        """
        Delete an equipment and all its associated sensors/readings.

        :param equipment_id: The equipment_id to delete
        :return: dict with 'result'
        """
        token = self.get_token()
        result = requests.delete(
            self.url + "/equipment/",
            json={"equipment_id": equipment_id},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    # ------------------------------------------------------------------ #
    #  Connector Management (CRUD)                                        #
    # ------------------------------------------------------------------ #

    def get_connectors(self, facility_name):
        """
        List all connectors for a facility.
        Note: config values are masked as '*****' by the backend for security.

        :param facility_name: Name of the facility
        :return: list of connector dicts
        """
        token = self.get_token()
        result = requests.post(
            self.url + "/connector/",
            json={"facility_name": facility_name},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    def create_connector(self, facility_name, name, connector_type=None, config=None):
        """
        Create a new connector for third-party data integration.

        :param facility_name: Facility name
        :param name: Connector display name
        :param connector_type: Type (e.g. "bacnet", "modbus", "mqtt", "http")
        :param config: JSON dict with connection credentials/settings
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "name": name,
        }
        if connector_type is not None:
            payload["type"] = connector_type
        if config is not None:
            payload["config"] = config
        result = requests.put(
            self.url + "/connector/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def update_connector(self, connector_id, facility_name, name, connector_type=None, config=None):
        """
        Update an existing connector.
        Pass config values as '*****' to preserve existing secret values.

        :param connector_id: The connector_id to update
        :param facility_name: Facility name
        :param name: Connector name
        :param connector_type: Connector type
        :param config: JSON config dict
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "connector_id": connector_id,
            "name": name,
        }
        if connector_type is not None:
            payload["type"] = connector_type
        if config is not None:
            payload["config"] = config
        result = requests.put(
            self.url + "/connector/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def delete_connector(self, connector_id):
        """
        Delete a connector.

        :param connector_id: The connector_id to delete
        :return: dict with 'result'
        """
        token = self.get_token()
        result = requests.delete(
            self.url + "/connector/",
            json={"connector_id": connector_id},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    # ------------------------------------------------------------------ #
    #  Equipment Relationship Management (CRUD)                           #
    # ------------------------------------------------------------------ #

    def get_equipment_relationships(self, facility_name):
        """
        List all equipment relationships for a facility.
        Relationships define parent-child or served-by links between equipment
        (e.g. chiller serves AHU, pump belongs to chiller plant).

        :param facility_name: Name of the facility
        :return: list of equipment relationship dicts
        """
        token = self.get_token()
        result = requests.post(
            self.url + "/equipment-relationship/",
            json={"facility_name": facility_name},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    def create_equipment_relationship(self, facility_name, equip1, equip2, relationship_type):
        """
        Create a relationship between two equipment items.

        :param facility_name: Facility name
        :param equip1: First equipment name (e.g. parent / server)
        :param equip2: Second equipment name (e.g. child / served)
        :param relationship_type: Relationship type string (e.g. "serves", "contains")
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "equip1": equip1,
            "equip2": equip2,
            "type": relationship_type,
        }
        result = requests.put(
            self.url + "/equipment-relationship/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def update_equipment_relationship(self, equipment_relationship_id, facility_name,
                                      equip1, equip2, relationship_type):
        """
        Update an existing equipment relationship.

        :param equipment_relationship_id: The relationship ID to update
        :param facility_name: Facility name
        :param equip1: First equipment name
        :param equip2: Second equipment name
        :param relationship_type: Relationship type string
        :return: dict with 'result' and 'record'
        """
        token = self.get_token()
        payload = {
            "facility_name": facility_name,
            "equipment_relationship_id": equipment_relationship_id,
            "equip1": equip1,
            "equip2": equip2,
            "type": relationship_type,
        }
        result = requests.put(
            self.url + "/equipment-relationship/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def delete_equipment_relationship(self, equipment_relationship_id):
        """
        Delete an equipment relationship.

        :param equipment_relationship_id: The relationship ID to delete
        :return: dict with 'result'
        """
        token = self.get_token()
        result = requests.delete(
            self.url + "/equipment-relationship/",
            json={"equipment_relationship_id": equipment_relationship_id},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    # ------------------------------------------------------------------ #
    #  Data Ingestion                                                     #
    # ------------------------------------------------------------------ #

    def upload_series(self, sensor_id, timestamps, values):
        """
        Upload time-series data for a single sensor.

        :param sensor_id: The sensor_id to upload data for
        :param timestamps: List of timestamp strings (ISO format)
        :param values: List of numeric values
        :return: server response dict
        """
        token = self.get_token()
        result = requests.put(
            self.url + "/series-upload/",
            json={
                "sensor_id": sensor_id,
                "timestamps": timestamps,
                "values": values,
            },
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def get_latest_readings(self, facility_name, equipment_name=None):
        """
        Get the latest readings for sensors in a facility.

        :param facility_name: Facility name
        :param equipment_name: Optional equipment filter
        :return: list of latest reading dicts
        """
        token = self.get_token()
        payload = {"facility_name": facility_name}
        if equipment_name is not None:
            payload["equipment_name"] = equipment_name
        result = requests.post(
            self.url + "/latest-readings/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["result"]

    # ------------------------------------------------------------------ #
    #  Custom View Management (CRUD)                                      #
    # ------------------------------------------------------------------ #

    def get_custom_views(self):
        """
        Get all custom views accessible to the current user.
        Admins see all views; other roles see only their own.

        :return: list of custom view dicts
        """
        token = self.get_token()
        result = requests.post(
            self.url + "/custom-view/",
            json={},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()["views"]

    def update_custom_view(self, view_id, view_json, published=False):
        """
        Create or update a custom view definition.

        :param view_id: Unique identifier for the view
        :param view_json: JSON dict containing the view definition
        :param published: Whether the view is published (default False)
        :return: dict with 'result'
        """
        token = self.get_token()
        payload = {
            "view_id": view_id,
            "json": view_json,
            "published": published,
        }
        result = requests.put(
            self.url + "/custom-view/",
            json=payload,
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()

    def delete_custom_view(self, view_id):
        """
        Delete a custom view.

        :param view_id: The view_id to delete
        :return: dict with 'result'
        """
        token = self.get_token()
        result = requests.delete(
            self.url + "/custom-view/",
            json={"view_id": view_id},
            headers={"Authorization": "Bearer " + token, 'Content-Type': 'application/json'}
        )
        assert result.status_code == 200
        return result.json()



import datetime
import json
from unittest import mock
from unittest.mock import patch, MagicMock
from copy import deepcopy
from freezegun import freeze_time

import pandas as pd
import pytest

from carnotpy import process_equipment, process_date_range, CarnotPy


# ======================================================================
# Helper factories
# ======================================================================

def _mock_response(status_code=200, json_data=None, text=None):
    """Return a MagicMock that behaves like a requests.Response."""
    resp = MagicMock()
    resp.status_code = status_code
    resp.json.return_value = json_data or {}
    resp.text = text or json.dumps(json_data or {})
    return resp


def _make_client(**overrides):
    defaults = dict(
        url="https://api.example.com",
        username="user",
        password="pass",
    )
    defaults.update(overrides)
    return CarnotPy(**defaults)


# ======================================================================
# Tests for process_equipment()
# ======================================================================

class TestProcessEquipment:

    def test_replaces_equipment_in_filter_stage(self):
        payload = {
            "filters": {"equipment_name": "Chiller-1"},
            "pipeline": [
                {"type": "filter", "filter_by": "Chiller-1 - temperature"},
            ],
        }
        process_equipment(payload, "Chiller-2")
        assert payload["pipeline"][0]["filter_by"] == "Chiller-2 - temperature"
        assert payload["filters"]["equipment_name"] == "Chiller-2"

    def test_replaces_equipment_in_formula_stage(self):
        payload = {
            "filters": {"equipment_name": "AHU-1"},
            "pipeline": [
                {"type": "formula", "formula": "AHU-1 - flow * 2"},
            ],
        }
        process_equipment(payload, "AHU-3")
        assert payload["pipeline"][0]["formula"] == "AHU-3 - flow * 2"

    def test_replaces_equipment_in_insert_series_stage(self):
        payload = {
            "filters": {"equipment_name": "Pump-A"},
            "pipeline": [
                {
                    "type": "insert_series",
                    "name": "Pump-A - pressure",
                    "filters": {"equipment_name": "Pump-A"},
                },
            ],
        }
        process_equipment(payload, "Pump-B")
        assert payload["pipeline"][0]["name"] == "Pump-B - pressure"
        assert payload["pipeline"][0]["filters"]["equipment_name"] == "Pump-B"

    def test_insert_series_no_matching_filter(self):
        """insert_series whose sub-filter equipment_name differs from the top-level one."""
        payload = {
            "filters": {"equipment_name": "Pump-A"},
            "pipeline": [
                {
                    "type": "insert_series",
                    "name": "Pump-A - pressure",
                    "filters": {"equipment_name": "Other"},
                },
            ],
        }
        process_equipment(payload, "Pump-B")
        # Sub-filter should NOT be replaced because it doesn't match the top-level name
        assert payload["pipeline"][0]["filters"]["equipment_name"] == "Other"

    def test_noop_when_no_equipment_name_in_filters(self):
        payload = {"filters": {}, "pipeline": []}
        process_equipment(payload, "X")
        assert payload["filters"] == {}

    def test_noop_when_no_pipeline(self):
        payload = {"filters": {"equipment_name": "A"}}
        process_equipment(payload, "B")
        assert payload["filters"]["equipment_name"] == "B"

    def test_empty_pipeline(self):
        payload = {"filters": {"equipment_name": "A"}, "pipeline": []}
        process_equipment(payload, "B")
        assert payload["filters"]["equipment_name"] == "B"


# ======================================================================
# Tests for process_date_range()
# ======================================================================

@freeze_time("2025-06-15 14:30:00")
class TestProcessDateRangeRelative:
    """Custom relative ranges (useCustomRange + useCustomRelativeRange)."""

    def _make_payload(self, rel_range, **extra):
        p = {
            "useCustomRange": True,
            "useCustomRelativeRange": True,
            "customRelativeRange": rel_range,
        }
        p.update(extra)
        return p

    def test_current_week(self):
        payload = self._make_payload("current week")
        process_date_range(payload, None)
        # 2025-06-15 is a Sunday (weekday=6)
        assert payload["date_range"][0] == "2025-06-09"  # Monday

    def test_current_month(self):
        payload = self._make_payload("current month")
        process_date_range(payload, None)
        assert payload["date_range"][0] == "2025-06-01"

    def test_today(self):
        payload = self._make_payload("today")
        process_date_range(payload, None)
        assert payload["date_range"][0] == "2025-06-15 00:00:00"
        assert payload["date_range"][1] == "2025-06-15 14:30:00"

    def test_today_full(self):
        payload = self._make_payload("today full")
        process_date_range(payload, None)
        # 'today full' is not in the explicit time-format list, so date-only start
        assert payload["date_range"][0] == "2025-06-15"

    def test_current_year(self):
        payload = self._make_payload("current year")
        process_date_range(payload, None)
        assert payload["date_range"][0] == "2025-01-01"

    def test_last_week(self):
        payload = self._make_payload("last week")
        process_date_range(payload, None)
        # last week: base moves back 1 week to 2025-06-08 (Sun)
        # start_date = end_date - weekday offset → Monday of that week
        start = payload["date_range"][0]
        assert "2025-06" in start

    def test_last_month(self):
        payload = self._make_payload("last month")
        process_date_range(payload, None)
        assert payload["date_range"][0] == "2025-05-01"

    def test_yesterday(self):
        payload = self._make_payload("yesterday")
        process_date_range(payload, None)
        assert payload["date_range"][0] == "2025-06-14 00:00:00"
        # end uses base_date (which keeps the original frozen time component)
        assert payload["date_range"][1] == "2025-06-14 14:30:00"

    def test_last_year(self):
        payload = self._make_payload("last year")
        process_date_range(payload, None)
        # Note: source code does year-1 twice for start (base_date already decremented)
        assert payload["date_range"][0].startswith("2023-01-01")
        assert payload["date_range"][1].startswith("2024-12-31")

    def test_today_and_tomorrow(self):
        payload = self._make_payload("today and tomorrow")
        process_date_range(payload, None)
        assert payload["date_range"][0] == "2025-06-15 00:00:00"
        assert payload["date_range"][1] == "2025-06-16 23:59:59"

    def test_default_30_days(self):
        payload = self._make_payload("30 days")
        process_date_range(payload, None)
        assert payload["date_range"][0] == "2025-05-16"

    def test_custom_relative_start_only(self):
        payload = self._make_payload("today", customRelativeRangeStartOnly=True)
        date_range = {"start": "2025-01-01", "end": "2025-12-31"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2025-06-15 00:00:00"
        assert payload["date_range"][1] == "2025-12-31 23:59:59"


class TestProcessDateRangeCustom:
    """Custom fixed ranges (useCustomRange + customRange)."""

    def test_custom_range(self):
        payload = {
            "useCustomRange": True,
            "customRange": {"start": "2025-03-01", "end": "2025-03-31"},
        }
        process_date_range(payload, None)
        assert payload["date_range"] == ["2025-03-01", "2025-03-31 23:59:59"]

    def test_custom_range_start_only(self):
        payload = {
            "useCustomRange": True,
            "customRange": {"start": "2025-03-01", "end": "2025-03-31"},
            "customRangeStartOnly": True,
        }
        date_range = {"start": "2025-01-01", "end": "2025-06-30"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2025-03-01"
        assert payload["date_range"][1] == "2025-06-30 23:59:59"

    def test_custom_range_end_already_has_time(self):
        payload = {
            "useCustomRange": True,
            "customRange": {"start": "2025-03-01", "end": "2025-03-31 12:00:00"},
        }
        process_date_range(payload, None)
        assert payload["date_range"][1] == "2025-03-31 12:00:00"


class TestProcessDateRangeOffset:
    """Offset date ranges (no useCustomRange, using offsetRange)."""

    def test_offset_days(self):
        payload = {
            "offsetRange": True,
            "offsetType": "days",
            "offsetAmount": 7,
            "offsetRangeStartOnly": False,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2025-06-08"
        assert payload["date_range"][1] == "2025-06-15 23:59:59"

    def test_offset_months(self):
        payload = {
            "offsetRange": True,
            "offsetType": "months",
            "offsetAmount": 1,
            "offsetRangeStartOnly": False,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2025-05-15"
        assert payload["date_range"][1] == "2025-05-22 23:59:59"

    def test_offset_years(self):
        payload = {
            "offsetRange": True,
            "offsetType": "years",
            "offsetAmount": 1,
            "offsetRangeStartOnly": False,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2024-06-15"
        assert payload["date_range"][1] == "2024-06-22 23:59:59"

    def test_offset_precise_year(self):
        payload = {
            "offsetRange": True,
            "offsetType": "years",
            "offsetAmount": 1,
            "offsetPreciseYear": 2023,
            "offsetRangeStartOnly": False,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2023-06-15"
        assert payload["date_range"][1] == "2023-06-22 23:59:59"

    def test_offset_start_only_days(self):
        payload = {
            "offsetRange": True,
            "offsetType": "days",
            "offsetAmount": 7,
            "offsetRangeStartOnly": True,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2025-06-08"
        assert payload["date_range"][1] == "2025-06-22 23:59:59"

    def test_offset_start_only_months(self):
        payload = {
            "offsetRange": True,
            "offsetType": "months",
            "offsetAmount": 2,
            "offsetRangeStartOnly": True,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2025-04-15"
        assert payload["date_range"][1] == "2025-06-22 23:59:59"

    def test_offset_start_only_years(self):
        payload = {
            "offsetRange": True,
            "offsetType": "years",
            "offsetAmount": 2,
            "offsetRangeStartOnly": True,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2023-06-15"
        assert payload["date_range"][1] == "2025-06-22 23:59:59"

    def test_offset_start_only_precise_year(self):
        payload = {
            "offsetRange": True,
            "offsetType": "year",
            "offsetAmount": 1,
            "offsetPreciseYear": 2020,
            "offsetRangeStartOnly": True,
        }
        date_range = {"start": "2025-06-15", "end": "2025-06-22"}
        process_date_range(payload, date_range)
        assert payload["date_range"][0] == "2020-06-15"
        assert payload["date_range"][1] == "2025-06-22 23:59:59"


class TestProcessDateRangePassthrough:
    """No custom range, no offset → simple passthrough."""

    def test_passthrough(self):
        payload = {}
        date_range = {"start": "2025-01-01", "end": "2025-12-31"}
        process_date_range(payload, date_range)
        assert payload["date_range"] == ["2025-01-01", "2025-12-31 23:59:59"]

    def test_passthrough_end_has_time(self):
        payload = {}
        date_range = {"start": "2025-01-01", "end": "2025-12-31 11:00:00"}
        process_date_range(payload, date_range)
        assert payload["date_range"] == ["2025-01-01", "2025-12-31 11:00:00"]

    def test_no_date_range(self):
        payload = {}
        process_date_range(payload, None)
        assert "date_range" not in payload


# ======================================================================
# Tests for CarnotPy class
# ======================================================================

class TestCarnotPyInit:
    def test_init(self):
        c = _make_client()
        assert c.url == "https://api.example.com"
        assert c.username == "user"
        assert c.password == "pass"
        assert c.preset_token is None

    def test_init_with_preset_token(self):
        c = _make_client(preset_token="tok123")
        assert c.preset_token == "tok123"


class TestGetToken:

    @patch("carnotpy.requests.post")
    def test_returns_token_from_login(self, mock_post):
        mock_post.return_value = _mock_response(200, {"token": "abc"})
        c = _make_client()
        assert c.get_token() == "abc"
        mock_post.assert_called_once_with(
            "https://api.example.com/login/",
            json={"user_name": "user", "password": "pass"},
        )

    def test_returns_preset_token(self):
        c = _make_client(preset_token="preset")
        assert c.get_token() == "preset"

    @patch("carnotpy.requests.post")
    def test_raises_on_invalid_credentials(self, mock_post):
        mock_post.return_value = _mock_response(200, {"error": "bad"})
        c = _make_client()
        with pytest.raises(ValueError, match="Invalid username or password"):
            c.get_token()

    @patch("carnotpy.requests.post")
    def test_raises_on_non_200(self, mock_post):
        mock_post.return_value = _mock_response(401)
        c = _make_client()
        with pytest.raises(AssertionError):
            c.get_token()


class TestDiscover:

    @patch("carnotpy.requests.post")
    def test_discover(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": [{"id": 1}]}),
        ]
        c = _make_client()
        assert c.discover() == [{"id": 1}]
        assert mock_post.call_count == 2


class TestGetNotifications:

    @patch("carnotpy.requests.post")
    def test_get_notifications(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": [{"msg": "alert"}]}),
        ]
        c = _make_client()
        res = c.get_notifications("facility1", date_range=["2025-01-01", "2025-12-31"])
        assert res == [{"msg": "alert"}]


class TestGetConfig:

    @patch("carnotpy.requests.post")
    def test_get_config(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": {"key": "value"}}),
        ]
        c = _make_client()
        assert c.get_config("fac") == {"key": "value"}


class TestSetConfig:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_set_config(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "ok"})
        c = _make_client()
        res = c.set_config("fac", "cfg_name", {"a": 1}, config_id="cid")
        assert res == {"result": "ok"}
        call_json = mock_put.call_args[1]["json"]
        assert call_json["facility_name"] == "fac"
        assert call_json["config_name"] == "cfg_name"
        assert call_json["config_id"] == "cid"


class TestGetMaintenanceRecords:

    @patch("carnotpy.requests.post")
    def test_get_maintenance_records(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": []}),
        ]
        c = _make_client()
        assert c.get_maintenance_records("fac") == []


class TestReadHistory:

    @patch("carnotpy.requests.post")
    def test_returns_dataframe(self, mock_post):
        data = {
            "timestamp": ["2025-01-01T00:00:00", "2025-01-01T01:00:00"],
            "value": [10, 20],
        }
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, data),
        ]
        c = _make_client()
        df = c.read_history(["s1"], "2025-01-01", "2025-01-02")
        assert isinstance(df, pd.DataFrame)
        assert df.index.name == "timestamp"
        assert list(df["value"]) == [10, 20]


class TestReadData3:

    @patch("carnotpy.requests.post")
    def test_returns_dataframe(self, mock_post):
        data = {
            "timestamp": ["2025-01-01T00:00:00", "2025-01-01T01:00:00"],
            "col1": [1.0, 2.0],
        }
        mock_post.side_effect = [
            _mock_response(200, data, text=json.dumps(data)),
        ]
        c = _make_client()
        df = c.read_data3({"some": "payload"}, token="pre")
        assert isinstance(df, pd.DataFrame)
        assert list(df.columns) == ["col1"]


class TestReadData3Query:

    @patch("carnotpy.requests.post")
    def test_basic_query(self, mock_post):
        data = {
            "timestamp": ["2025-01-01T00:00:00"],
            "val": [42],
        }
        login_resp = _mock_response(200, {"token": "t"})
        data_resp = _mock_response(200, data, text=json.dumps(data))
        mock_post.side_effect = [login_resp, data_resp]

        query = {
            "data-payload": [
                {
                    "filters": {"facility_name": "fac"},
                    "seriesOverrides": [],
                }
            ]
        }
        c = _make_client()
        df = c.read_data3_query(
            query, "fac", {"start": "2025-01-01", "end": "2025-01-02"}
        )
        assert isinstance(df, pd.DataFrame)

    @patch("carnotpy.requests.post")
    def test_query_with_equipment_and_on_only(self, mock_post):
        data = {
            "timestamp": ["2025-01-01T00:00:00"],
            "val": [1],
        }
        login_resp = _mock_response(200, {"token": "t"})
        data_resp = _mock_response(200, data, text=json.dumps(data))
        mock_post.side_effect = [login_resp, data_resp]

        query = {
            "data-payload": [
                {
                    "filters": {"equipment_name": "Chiller-1"},
                    "seriesOverrides": [],
                }
            ]
        }
        c = _make_client()
        df = c.read_data3_query(
            query, "fac", {"start": "2025-01-01", "end": "2025-01-02"},
            equipment="Chiller-2", on_only=True,
        )
        assert isinstance(df, pd.DataFrame)

    @patch("carnotpy.requests.post")
    def test_independent_x_axis(self, mock_post):
        data1 = {"timestamp": ["2025-01-01T00:00:00"], "a": [1]}
        data2 = {"timestamp": ["2025-01-01T01:00:00"], "b": [2]}
        login_resp = _mock_response(200, {"token": "t"})
        resp1 = _mock_response(200, data1, text=json.dumps(data1))
        resp2 = _mock_response(200, data2, text=json.dumps(data2))
        mock_post.side_effect = [login_resp, resp1, resp2]

        query = {
            "data-payload": [
                {"filters": {"facility_name": "fac"}, "seriesOverrides": []},
                {"filters": {"facility_name": "fac"}, "seriesOverrides": []},
            ]
        }
        c = _make_client()
        df = c.read_data3_query(
            query, "fac", {"start": "2025-01-01", "end": "2025-01-02"},
            independent_x_axis=True,
        )
        assert "a" in df.columns
        assert "b" in df.columns


class TestGetSensorId:

    @patch("carnotpy.requests.post")
    def test_found(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": [
                {"sensor_name": "temp", "sensor_id": "s1"},
                {"sensor_name": "flow", "sensor_id": "s2"},
            ]}),
        ]
        c = _make_client()
        assert c.get_sensor_id("fac", "equip", "flow") == "s2"

    @patch("carnotpy.requests.post")
    def test_not_found(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": []}),
        ]
        c = _make_client()
        with pytest.raises(KeyError, match="Sensor not found"):
            c.get_sensor_id("fac", "equip", "missing")


class TestWrite:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_write(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "ok"})
        c = _make_client()
        res = c.write("s1", 72, 8, duration=60)
        assert res == {"result": "ok"}
        call_json = mock_put.call_args[1]["json"]
        assert call_json["sensor_id"] == "s1"
        assert call_json["value"] == 72
        assert call_json["priority"] == 8
        assert call_json["duration"] == 60


class TestAddJob:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_add_job(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "queued"})
        c = _make_client()
        res = c.add_job("fac", "export", {"format": "csv"})
        assert res == {"result": "queued"}


# ------------------------------------------------------------------ #
#  Point CRUD
# ------------------------------------------------------------------ #

class TestGetPoints:

    @patch("carnotpy.requests.post")
    def test_get_points(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": [{"sensor_id": "s1"}]}),
        ]
        c = _make_client()
        assert c.get_points("fac") == [{"sensor_id": "s1"}]

    @patch("carnotpy.requests.post")
    def test_get_points_with_equipment(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": []}),
        ]
        c = _make_client()
        res = c.get_points("fac", equipment_name="equip")
        assert res == []
        call_json = mock_post.call_args[1]["json"]
        assert call_json["equipment_name"] == "equip"


class TestCreatePoint:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_minimal(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "created", "record": {}})
        c = _make_client()
        res = c.create_point("fac", "equip", "temp_sensor")
        assert res["result"] == "created"
        call_json = mock_put.call_args[1]["json"]
        assert call_json["sensor_name"] == "temp_sensor"
        assert call_json["faulty"] is False
        assert "unit" not in call_json

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_all_optional_fields(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "created", "record": {}})
        c = _make_client()
        c.create_point(
            "fac", "equip", "temp",
            sensor_name_abbr="T",
            sensor_desc="Temperature",
            unit="degC",
            haystack_id="h1",
            calculations_statement="df[2]+df[3]",
            weight=2,
            connector_id="c1",
            faulty=True,
            meta={"tag": "hot"},
        )
        call_json = mock_put.call_args[1]["json"]
        assert call_json["unit"] == "degC"
        assert call_json["meta"] == {"tag": "hot"}
        assert call_json["faulty"] is True


class TestUpdatePoint:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_updates_allowed_fields(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "updated", "record": {}})
        c = _make_client()
        c.update_point("s1", "fac", "equip", sensor_name="new_name", unit="kW")
        call_json = mock_put.call_args[1]["json"]
        assert call_json["sensor_name"] == "new_name"
        assert call_json["unit"] == "kW"

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_ignores_disallowed_fields(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "updated", "record": {}})
        c = _make_client()
        c.update_point("s1", "fac", "equip", unknown_field="should_be_ignored")
        call_json = mock_put.call_args[1]["json"]
        assert "unknown_field" not in call_json


class TestDeletePoint:

    @patch("carnotpy.requests.delete")
    @patch("carnotpy.requests.post")
    def test_delete(self, mock_post, mock_delete):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_delete.return_value = _mock_response(200, {"result": "deleted"})
        c = _make_client()
        assert c.delete_point("s1") == {"result": "deleted"}


# ------------------------------------------------------------------ #
#  Equipment CRUD
# ------------------------------------------------------------------ #

class TestGetEquipment:

    @patch("carnotpy.requests.post")
    def test_get_equipment(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": [{"equipment_name": "AHU-1"}]}),
        ]
        c = _make_client()
        assert c.get_equipment("fac") == [{"equipment_name": "AHU-1"}]


class TestCreateEquipment:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_minimal(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "created", "record": {}})
        c = _make_client()
        res = c.create_equipment("fac", "AHU-1")
        assert res["result"] == "created"

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_with_all_optional(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "ok"})
        c = _make_client()
        c.create_equipment(
            "fac", "AHU-1",
            equipment_type="ahu",
            equipment_tags=["hvac"],
            floor="2F",
            notes="north wing",
            meta={"brand": "Acme"},
            graphic_id="g1",
        )
        call_json = mock_put.call_args[1]["json"]
        assert call_json["equipment_type"] == "ahu"
        assert call_json["graphic_id"] == "g1"


class TestUpdateEquipment:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_update(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "updated"})
        c = _make_client()
        c.update_equipment("e1", "fac", "AHU-1", floor="3F")
        call_json = mock_put.call_args[1]["json"]
        assert call_json["floor"] == "3F"
        assert call_json["equipment_id"] == "e1"

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_ignores_disallowed(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "updated"})
        c = _make_client()
        c.update_equipment("e1", "fac", "AHU-1", bad_field="x")
        call_json = mock_put.call_args[1]["json"]
        assert "bad_field" not in call_json


class TestDeleteEquipment:

    @patch("carnotpy.requests.delete")
    @patch("carnotpy.requests.post")
    def test_delete(self, mock_post, mock_delete):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_delete.return_value = _mock_response(200, {"result": "deleted"})
        c = _make_client()
        assert c.delete_equipment("e1") == {"result": "deleted"}


# ------------------------------------------------------------------ #
#  Connector CRUD
# ------------------------------------------------------------------ #

class TestGetConnectors:

    @patch("carnotpy.requests.post")
    def test_get(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": []}),
        ]
        c = _make_client()
        assert c.get_connectors("fac") == []


class TestCreateConnector:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_create(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "created"})
        c = _make_client()
        res = c.create_connector("fac", "My BACnet", connector_type="bacnet", config={"ip": "10.0.0.1"})
        assert res["result"] == "created"
        call_json = mock_put.call_args[1]["json"]
        assert call_json["type"] == "bacnet"


class TestUpdateConnector:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_update(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "updated"})
        c = _make_client()
        c.update_connector("c1", "fac", "My BACnet", connector_type="modbus")
        call_json = mock_put.call_args[1]["json"]
        assert call_json["connector_id"] == "c1"
        assert call_json["type"] == "modbus"


class TestDeleteConnector:

    @patch("carnotpy.requests.delete")
    @patch("carnotpy.requests.post")
    def test_delete(self, mock_post, mock_delete):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_delete.return_value = _mock_response(200, {"result": "deleted"})
        c = _make_client()
        assert c.delete_connector("c1") == {"result": "deleted"}


# ------------------------------------------------------------------ #
#  Equipment Relationship CRUD
# ------------------------------------------------------------------ #

class TestGetEquipmentRelationships:

    @patch("carnotpy.requests.post")
    def test_get(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": [{"equip1": "A", "equip2": "B"}]}),
        ]
        c = _make_client()
        res = c.get_equipment_relationships("fac")
        assert len(res) == 1


class TestCreateEquipmentRelationship:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_create(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "created"})
        c = _make_client()
        res = c.create_equipment_relationship("fac", "Chiller-1", "AHU-1", "serves")
        assert res["result"] == "created"
        call_json = mock_put.call_args[1]["json"]
        assert call_json["type"] == "serves"


class TestUpdateEquipmentRelationship:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_update(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "updated"})
        c = _make_client()
        res = c.update_equipment_relationship("r1", "fac", "A", "B", "contains")
        assert res["result"] == "updated"
        call_json = mock_put.call_args[1]["json"]
        assert call_json["equipment_relationship_id"] == "r1"


class TestDeleteEquipmentRelationship:

    @patch("carnotpy.requests.delete")
    @patch("carnotpy.requests.post")
    def test_delete(self, mock_post, mock_delete):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_delete.return_value = _mock_response(200, {"result": "deleted"})
        c = _make_client()
        assert c.delete_equipment_relationship("r1") == {"result": "deleted"}


# ------------------------------------------------------------------ #
#  Data Ingestion
# ------------------------------------------------------------------ #

class TestUploadSeries:

    @patch("carnotpy.requests.put")
    @patch("carnotpy.requests.post")
    def test_upload(self, mock_post, mock_put):
        mock_post.return_value = _mock_response(200, {"token": "t"})
        mock_put.return_value = _mock_response(200, {"result": "ok"})
        c = _make_client()
        res = c.upload_series("s1", ["2025-01-01T00:00:00"], [42])
        assert res["result"] == "ok"
        call_json = mock_put.call_args[1]["json"]
        assert call_json["sensor_id"] == "s1"
        assert call_json["timestamps"] == ["2025-01-01T00:00:00"]


class TestGetLatestReadings:

    @patch("carnotpy.requests.post")
    def test_get(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": [{"sensor_id": "s1", "value": 72}]}),
        ]
        c = _make_client()
        res = c.get_latest_readings("fac")
        assert res[0]["value"] == 72

    @patch("carnotpy.requests.post")
    def test_with_equipment(self, mock_post):
        mock_post.side_effect = [
            _mock_response(200, {"token": "t"}),
            _mock_response(200, {"result": []}),
        ]
        c = _make_client()
        c.get_latest_readings("fac", equipment_name="AHU-1")
        call_json = mock_post.call_args[1]["json"]
        assert call_json["equipment_name"] == "AHU-1"
